# -*- coding: utf-8 -*-
import re
from shutil import rmtree
import unittest

from skoolkittest import SkoolKitTestCase
from skoolkit import VERSION, SkoolKitError
from skoolkit.skoolhtml import HtmlWriter
from skoolkit.skoolparser import CASE_UPPER, CASE_LOWER

ODIR = 'html'
TEST_REF = 'test-html.ref'
TEST_NO_REF = 'test-html-no-ref.skool'

USAGE = """Usage: skool2html.py [options] FILE [FILE...]

  Convert skool files and ref files to HTML. FILE may be a regular file, or '-'
  for standard input.

Options:
  -V        Show SkoolKit version number and exit
  -q        Be quiet
  -t        Show timings
  -d DIR    Write files in this directory
  -o        Overwrite existing image files
  -l        Write disassembly in lower case
  -u        Write disassembly in upper case
  -D        Write disassembly in decimal
  -H        Write disassembly in hexadecimal
  -c S/L    Add the line 'L' to the ref file section 'S'; this option may be
            used multiple times
  -P PAGES  Write only these custom pages (when '-w P' is specified); PAGES
            should be a comma-separated list of IDs of pages defined in [Page:*]
            sections in the ref file(s)
  -w X      Write only these files, where X is one or more of:
              B = Graphic glitches
              b = Bugs
              c = Changelog
              d = Disassembly files
              G = Game status buffer
              g = Graphics
              i = Disassembly index
              m = Memory maps
              o = Other code
              P = Pages defined in the ref file(s)
              p = Pokes
              t = Trivia
              y = Glossary"""

OUTPUT_NO_REF = """Creating directory html
Using skool file: test-html-no-ref.skool
Found no ref file for test-html-no-ref.skool
Parsing test-html-no-ref.skool
Creating directory html/test-html-no-ref
Copying skoolkit.css to html/test-html-no-ref/skoolkit.css
  Writing disassembly files in test-html-no-ref/asm
  Writing test-html-no-ref/maps/all.html
  Writing test-html-no-ref/maps/routines.html
  Writing test-html-no-ref/maps/data.html
  Writing test-html-no-ref/maps/messages.html
  Writing test-html-no-ref/index.html"""

html_writer = None

class TestHtmlWriter(HtmlWriter):
    def init(self):
        global html_writer
        html_writer = self
        self.call_dict = {}

    def add_call(self, method_name, args):
        self.call_dict.setdefault(method_name, []).append(args)

    def write_logo_image(self, *args):
        self.add_call('write_logo_image', args)
        return True

    def write_asm_entries(self, *args):
        self.add_call('write_asm_entries', args)

    def write_map(self, *args):
        self.add_call('write_map', args)

    def write_page(self, *args):
        self.add_call('write_page', args)

    def write_gbuffer(self, *args):
        self.add_call('write_gbuffer', args)

    def write_graphics(self, *args):
        self.add_call('write_graphics', args)

    def write_graphic_glitches(self, *args):
        self.add_call('write_graphic_glitches', args)

    def write_changelog(self, *args):
        self.add_call('write_changelog', args)

    def write_bugs(self, *args):
        self.add_call('write_bugs', args)

    def write_facts(self, *args):
        self.add_call('write_facts', args)

    def write_glossary(self, *args):
        self.add_call('write_glossary', args)

    def write_pokes(self, *args):
        self.add_call('write_pokes', args)

    def write_entries(self, *args):
        self.add_call('write_entries', args)

    def write_index(self, *args):
        self.add_call('write_index', args)

class Skool2HtmlTest(SkoolKitTestCase):
    def tearDown(self):
        rmtree(ODIR, True)
        SkoolKitTestCase.tearDown(self)

    def test_no_arguments(self):
        output, error = self.run_skool2html(err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'))

    def test_invalid_option(self):
        output, error = self.run_skool2html('-X', err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'))

    def test_no_ref(self):
        output, error = self.run_skool2html('-d {0} {1}'.format(ODIR, TEST_NO_REF))
        self.assertEqual(len(error), 0)
        self.assert_output_equal(output, OUTPUT_NO_REF.split('\n'), True)

    def test_nonexistent_skool_file(self):
        skoolfile = 'xyz.skool'
        try:
            self.run_skool2html('-d {0} {1}'.format(ODIR, skoolfile))
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], '{0}: skool file not found'.format(skoolfile))

    def test_nonexistent_ref_file(self):
        reffile = 'zyx.ref'
        try:
            self.run_skool2html('-d {0} {1}'.format(ODIR, reffile))
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], '{0}: ref file not found'.format(reffile))

    def test_nonexistent_css_file(self):
        cssfile = 'abc.css'
        try:
            self.run_skool2html('-c Paths/StyleSheet={0} -w "" -d {1} {2}'.format(cssfile, ODIR, TEST_REF))
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], '{0}: file not found'.format(cssfile))

    def test_nonexistent_js_file(self):
        jsfile = 'cba.js'
        try:
            self.run_skool2html('-c Paths/JavaScript={0} -w "" -d {1} {2}'.format(jsfile, ODIR, TEST_REF))
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], '{0}: file not found'.format(jsfile))

    def test_invalid_page_id(self):
        page_id = 'NonexistentPage'
        try:
            self.run_skool2html('-d {0} -w P -P {1} {2}'.format(ODIR, page_id, TEST_REF))
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], 'Invalid page ID: {0}'.format(page_id))

    def test_option_w(self):
        options = [
            ('d', 'write_asm_entries', [()]),
            ('m', 'write_map', [({'Name': 'MemoryMap', 'PageByteColumns': '1'},),
                                ({'EntryTypes': 'c', 'Name': 'RoutinesMap'},),
                                ({'EntryTypes': 'bw', 'Name': 'DataMap', 'PageByteColumns': '1'},),
                                ({'EntryTypes': 't', 'Name': 'MessagesMap'},)]),
            ('P', 'write_page', [('CustomPage1',), ('CustomPage2',)]),
            ('G', 'write_gbuffer', [()]),
            ('g', 'write_graphics', [()]),
            ('B', 'write_graphic_glitches', [()]),
            ('c', 'write_changelog', [()]),
            ('b', 'write_bugs', [()]),
            ('t', 'write_facts', [()]),
            ('y', 'write_glossary', [()]),
            ('p', 'write_pokes', [()]),
            ('o', 'write_map', [({'Path': 'other.html', 'Title': 'Other code', 'AsmPath': 'other'},)]),
            ('o', 'write_entries', [('other', 'other.html', 'Other code')]),
            ('i', 'write_index', [()])
        ]
        for file_ids, method_name, exp_arg_list in options:
            output, error = self.run_skool2html('-d {0} -w {1} {2}'.format(ODIR, file_ids, TEST_REF))
            self.assertEqual(error, '')
            self.assertTrue(method_name in html_writer.call_dict, '{0} was not called'.format(method_name))
            arg_list = html_writer.call_dict[method_name]
            self.assertEqual(arg_list, exp_arg_list, '{0}: {1} != {2}'.format(method_name, arg_list, exp_arg_list))

    def test_option_V(self):
        output, error = self.run_skool2html('-V')
        self.assertEqual(error, '')
        self.assertEqual(len(output), 1)
        self.assertEqual(output[0], 'SkoolKit {0}'.format(VERSION))

    def test_option_q(self):
        output, error = self.run_skool2html('-q -d {0} -w i {1}'.format(ODIR, TEST_REF))
        self.assertEqual(error, '')
        logo_method = 'write_logo_image'
        self.assertTrue(logo_method in html_writer.call_dict, '{0} was not called'.format(logo_method))
        self.assertEqual(html_writer.call_dict[logo_method], [('{0}/test-html'.format(ODIR),)])
        self.assertEqual(len(output), 0)

    def test_option_t(self):
        output, error = self.run_skool2html('-t -w i -d {0} {1}'.format(ODIR, TEST_REF))
        self.assertEqual(error, '')
        pattern = 'Done \([0-9]+\.[0-9][0-9]s\)'
        done = output[-1]
        search = re.search(pattern, done)
        self.assertTrue(search is not None, '"{0}" is not of the form "{1}"'.format(done, pattern))

    def test_option_c(self):
        for section_name, param_name, value in (
            ('Colours', 'RED', '255,0,0'),
            ('Config', 'GameDir', 'test-c'),
            ('ImageWriter', 'DefaultFormat', 'gif')
        ):
            output, error = self.run_skool2html('-c {0}/{1}={2} -w i -d {3} {4}'.format(section_name, param_name, value, ODIR, TEST_REF))
            self.assertEqual(error, '')
            section = html_writer.ref_parser.get_dictionary(section_name)
            self.assertEqual(section[param_name], value, '{0}/{1}!={2}'.format(section_name, param_name, value))
        sl_spec = 'GARBAGE'
        try:
            self.run_skool2html('-c {0} -w i -d {1} {2}'.format(sl_spec, ODIR, TEST_REF))
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], 'Malformed SectionName/Line spec: {0}'.format(sl_spec))

    def test_option_o(self):
        output, error = self.run_skool2html('-o -w "" -d {0} {1}'.format(ODIR, TEST_REF))
        self.assertEqual(error, '')
        self.assertTrue(html_writer.file_info.replace_images)

    def test_option_l(self):
        output, error = self.run_skool2html('-l -w "" -d {0} {1}'.format(ODIR, TEST_REF))
        self.assertEqual(error, '')
        self.assertEqual(html_writer.case, CASE_LOWER)

    def test_option_u(self):
        output, error = self.run_skool2html('-u -w "" -d {0} {1}'.format(ODIR, TEST_REF))
        self.assertEqual(error, '')
        self.assertEqual(html_writer.case, CASE_UPPER)

    def test_option_D(self):
        output, error = self.run_skool2html('-D -w "" -d {0} {1}'.format(ODIR, TEST_REF))
        self.assertEqual(error, '')
        self.assertTrue(html_writer.parser.mode.decimal)

    def test_option_H(self):
        output, error = self.run_skool2html('-H -w "" -d {0} {1}'.format(ODIR, TEST_REF))
        self.assertEqual(error, '')
        self.assertTrue(html_writer.parser.mode.hexadecimal)

if __name__ == '__main__':
    unittest.main()
