# -*- coding: utf-8 -*-
import re
import unittest

from skoolkittest import PY3, SkoolKitTestCase
from skoolkit import SkoolParsingError
from skoolkit.skoolasm import AsmWriter
from skoolkit.skoolparser import SkoolParser, CASE_LOWER

TEST_SKOOL = 'test-asm.skool'

TABLE_SRC = []
TABLE_TEXT = []

# Headers, colspans 1 and 2, rowspans 1 and 2, transparent cell
TABLE_SRC.append("""(data)
{ =h Col1 | =h Col2 | =h,c2 Cols3+4 }
{ =r2 X   | Y       | Za  | Zb }
{           Y2      | Za2 | =t }""")

TABLE_TEXT.append("""
+------+------+----------+
| Col1 | Col2 | Cols3+4  |
+------+------+-----+----+
| X    | Y    | Za  | Zb |
|      | Y2   | Za2 |    |
+------+------+-----+----+
""")

# Cell with rowspan > 1 in the middle column
TABLE_SRC.append("""
{ A1 cell | =r2 This is a cell with rowspan=2 | C1 cell }
{ A2 cell | C2 cell }""")

TABLE_TEXT.append("""
+---------+-------------------------------+---------+
| A1 cell | This is a cell with rowspan=2 | C1 cell |
| A2 cell |                               | C2 cell |
+---------+-------------------------------+---------+
""")

# Cell with rowspan > 1 in the rightmost column
TABLE_SRC.append("""(,,,:w)
{ A1 cell | B1 cell | =r2 This is a cell with rowspan=2 }
{ A2 cell | B2 cell }""")

TABLE_TEXT.append("""
+---------+---------+-------------------------------+
| A1 cell | B1 cell | This is a cell with rowspan=2 |
| A2 cell | B2 cell |                               |
+---------+---------+-------------------------------+
""")

def get_chr(code):
    return chr(code) if PY3 else unichr(code).encode('utf-8')

class AsmWriterTest(SkoolKitTestCase):
    def _get_writer(self, crlf=False, tab=False, lower=False, instr_width=23, warn=False):
        case = CASE_LOWER if lower else None
        skool_parser = SkoolParser(TEST_SKOOL, case=case, asm_mode=1)
        return AsmWriter(skool_parser, crlf, tab, lower, instr_width, warn)

    def _get_asm(self, crlf=False, tab=False, lower=False, instr_width=23, warn=False):
        writer = self._get_writer(crlf, tab, lower, instr_width, warn)
        writer.write()
        asm = self.out.getvalue().split('\n')[:-1]
        if warn:
            return asm, self.err.getvalue().split('\n')[:-1]
        return asm

    def _test_unsupported_macro(self, text):
        search = re.search('#[A-Z]+', text)
        macro = search.group()
        writer = self._get_writer()
        try:
            writer.expand(text)
            self.fail()
        except SkoolParsingError as e:
            self.assertEqual(e.args[0], 'Found unsupported macro: {0}'.format(macro))

    def _test_reference_macro(self, macro, def_link_text):
        writer = self._get_writer()
        for link_text in ('',  '(testing)', '(testing (nested) parentheses)'):
            for anchor in ('', '#name'):
                output = writer.expand('#{0}{1}{2}'.format(macro, anchor, link_text))
                self.assertEqual(output, link_text[1:-1] or def_link_text)

    def _test_call(self, x, y):
        return y * x

    def test_macro_bug(self):
        self._test_reference_macro('BUG', 'bug')

    def test_macro_call(self):
        writer = self._get_writer()
        writer.test_call = self._test_call
        output = writer.expand('#CALL:test_call(10,t)')
        self.assertEqual(output, 'tttttttttt')

    def test_macro_chr(self):
        writer = self._get_writer()

        output = writer.expand('#CHR169')
        self.assertEqual(output, get_chr(169))

        output = writer.expand('#CHR(163)1985')
        self.assertEqual(output, '{0}1985'.format(get_chr(163)))

    def test_macro_d(self):
        writer = self._get_writer()

        output = writer.expand('#D24581')
        self.assertEqual(output, 'Another routine')

        output = writer.expand('#D$6010')
        self.assertEqual(output, 'Yet another routine still')

    def test_macro_erefs(self):
        writer = self._get_writer()
        output = writer.expand('#EREFS24582')
        self.assertEqual(output, 'routines at 24586 and 24592')

    def test_macro_fact(self):
        self._test_reference_macro('FACT', 'fact')

    def test_macro_font(self):
        self._test_unsupported_macro('#FONT55584,96,,,1')

    def test_macro_html(self):
        writer = self._get_writer()
        delimiters = {
            '(': ')',
            '[': ']',
            '{': '}'
        }
        for text in('', 'See <a href="url">this</a>', 'A &gt; B'):
            for delim1 in '([{!@$%^*_-+|':
                delim2 = delimiters.get(delim1, delim1)
                output = writer.expand('#HTML{0}{1}{2}'.format(delim1, text, delim2))
                self.assertEqual(output, '')

    def test_macro_link(self):
        writer = self._get_writer()
        link_text = 'Testing'
        output = writer.expand('#LINK:PageID#anchor({0})'.format(link_text))
        self.assertEqual(output, link_text)

    def test_macro_name(self):
        writer = self._get_writer()
        output = writer.expand('#NAMEanchor')
        self.assertEqual(output, '')

    def test_macro_poke(self):
        self._test_reference_macro('POKE', 'poke')

    def test_macro_pokes(self):
        writer = self._get_writer()
        snapshot = writer.snapshot

        output = writer.expand('#POKES32768,255')
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768], 255)

        output = writer.expand('#POKES32768,254,10')
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768:32778], [254] * 10)

        output = writer.expand('#POKES32768,253,20,2')
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768:32808:2], [253] * 20)

        output = writer.expand('#POKES49152,1;49153,2')
        self.assertEqual(output, '')
        self.assertEqual(snapshot[49152:49154], [1, 2])

    def test_macro_pops(self):
        writer = self._get_writer()
        addr, byte = 49152, 128
        writer.snapshot[addr] = byte
        writer.push_snapshot('test')
        writer.snapshot[addr] = (byte + 127) % 256
        output = writer.expand('#POPS')
        self.assertEqual(output, '')
        self.assertEqual(writer.snapshot[addr], byte)

    def test_macro_pushs(self):
        writer = self._get_writer()
        addr, byte = 32768, 64
        writer.snapshot[addr] = byte
        output = writer.expand('#PUSHStest')
        self.assertEqual(output, '')
        self.assertEqual(writer.snapshot[addr], byte)
        writer.snapshot[addr] = (byte + 127) % 256
        writer.pop_snapshot()
        self.assertEqual(writer.snapshot[addr], byte)

    def test_macro_r(self):
        writer = self._get_writer()

        output = writer.expand('#R24576')
        self.assertEqual(output, 'DOSTUFF')

        link_text = 'Testing1'
        output = writer.expand('#R24576({0})'.format(link_text))
        self.assertEqual(output, link_text)

        output = writer.expand('#R24579')
        self.assertEqual(output, '6003')

        output = writer.expand('#R24580')
        self.assertEqual(output, '6004')

        output = writer.expand('#R32768@other')
        self.assertEqual(output, '32768')

        link_text = 'Testing2'
        output = writer.expand('#R32768@other#testing({0})'.format(link_text))
        self.assertEqual(output, link_text)

    def test_macro_refs(self):
        writer = self._get_writer()
        output = writer.expand('#REFS24581')
        self.assertEqual(output, 'routines at 24583, 24586 and 24592')

    def test_macro_reg(self):
        writer = self._get_writer()
        for reg in ("a", "b", "c", "d", "e", "h", "l", "i", "r", "ixl", "ixh", "iyl", "iyh", "b'", "c'", "d'", "e'", "h'", "l'", "bc", "de", "hl", "sp", "ix", "iy", "bc'", "de'", "hl'"):
            output = writer.expand('#REG{0}'.format(reg))
            self.assertEqual(output, reg.upper())

    def test_macro_scr(self):
        self._test_unsupported_macro('#SCR2(fname)')

    def test_macro_space(self):
        writer = self._get_writer()

        output = writer.expand('"#SPACE"')
        self.assertEqual(output, '" "')

        output = writer.expand('"#SPACE5"')
        self.assertEqual(output, '"     "')

        output = writer.expand('1#SPACE(3)1')
        self.assertEqual(output, '1   1')

    def test_macro_table(self):
        writer = self._get_writer()
        for src, text in zip(TABLE_SRC, TABLE_TEXT):
            table = '#TABLE{0}\nTABLE#'.format(src)
            output = '\n'.join(writer.format(table, 79))
            self.assertEqual(output, text.strip())

    def test_macro_udg(self):
        self._test_unsupported_macro('#UDG39144,6(safe_key)')

    def test_macro_udgarray(self):
        self._test_unsupported_macro('#UDGARRAY8,,,256;33008-33023(bubble)')

    def test_macro_udgtable(self):
        writer = self._get_writer()
        udgtable = '#UDGTABLE{0}\nUDGTABLE#'.format(TABLE_SRC[0])
        output = writer.format(udgtable, 79)
        self.assertEqual(output, [])

    def test_option_crlf(self):
        asm = self._get_asm(crlf=True)
        self.assertTrue(all(line.endswith('\r') for line in asm))

    def test_option_tab(self):
        asm = self._get_asm(tab=True)
        for line in asm:
            if line and line[0].isspace():
                self.assertEqual(line[0], '\t')

    def test_option_lower(self):
        asm = self._get_asm(lower=True)

        # Test that the ORG directive is in lower case
        self.assertEqual(asm[0], '  org 24576')

        # Test that labels are unaffected
        self.assertEqual(asm[8], 'DOSTUFF:')

        # Test that #REG macro arguments are in lower case
        self.assertEqual(asm[14], '  defb 123                ; a=0')

        # Test that #R macro arguments that resolve to a label are unaffected
        self.assertEqual(asm[15], '  defb 246                ; DOSTUFF')

    def test_option_instr_width(self):
        for width in (5, 10, 15, 20, 25, 30):
            self.clear_streams()
            asm = self._get_asm(instr_width=width)
            self.assertEqual(asm[14], '  {0} ; A=0'.format('DEFB 123'.ljust(width)))

    def test_option_warn(self):
        asm, err = self._get_asm(warn=True)

        # Test warning about #R macro with address that has no label
        self.assertTrue('WARNING: Could not convert address $6003 to label' in err)

    def test_registers(self):
        asm = self._get_asm()

        # Traditional
        self.assertEqual(asm[56], '; A Some value')
        self.assertEqual(asm[57], '; B Some other value')

        # With prefixes (right-justified to longest prefix)
        self.assertEqual(asm[64], ';  Input:A Some value')
        self.assertEqual(asm[65], ';        B Some other value')
        self.assertEqual(asm[66], '; Output:C The result')

    def test_empty_description(self):
        asm = self._get_asm()
        self.assertEqual(asm[69], '; Test an empty description')
        self.assertEqual(asm[70], '  RET')

if __name__ == '__main__':
    unittest.main()
