# -*- coding: utf-8 -*-
import random
from shutil import rmtree
from os.path import join
import unittest

from skoolkittest import SkoolKitTestCase, StringIO
from skoolkit import VERSION, SkoolKitError
from skoolkit.skoolhtml import HtmlWriter, FileInfo, Udg
from skoolkit.skoolparser import SkoolParser, Register
from skoolkit.refparser import RefParser

TEST_SKOOL = 'test-html.skool'
TEST_REF = 'test-html.ref'
ODIR = 'html'
GAMEDIR = 'test'
ASMDIR = 'asm'
GRAPHICS_DIR = 'graphics'
BUFFERS_DIR = 'buffers'
REFERENCE_DIR = 'reference'
UDGDIR = 'images/udgs'
TEMPLATES_DIR = 'templates'

HEADER = """<?xml version="1.0" encoding="utf-8" ?>
<!DOCTYPE html
    PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<title>test-html: {title}</title>
<link rel="stylesheet" type="text/css" href="{path}skoolkit.css" />{script}
</head>
{body}
<table class="header">
<tr>
<td class="headerLogo"><a class="link" href="{path}index.html">test-html</a></td>
<td class="headerText">{header}</td>
</tr>
</table>
"""

FOOTER = """<div class="footer">
<div class="release">Test HTML disassembly</div>
<div class="copyright">Me, 2012</div>
<div class="created">Created using <a class="link" href="http://pyskool.ca/?page_id=177">SkoolKit</a> {0}.</div>
</div>
</body>
</html>""".format(VERSION)

TABLE_SRC = """(data)
{ =h Col1 | =h Col2 | =h,c2 Cols3+4 }
{ =r2 X   | Y       | Za  | Zb }
{           Y2      | Za2 | =t }
"""

TABLE_HTML = """<table class="data">
<tr>
<th>Col1</th>
<th>Col2</th>
<th colspan="2">Cols3+4</th>
</tr>
<tr>
<td rowspan="2">X</td>
<td>Y</td>
<td>Za</td>
<td>Zb</td>
</tr>
<tr>
<td>Y2</td>
<td>Za2</td>
<td class="transparent"></td>
</tr>
</table>"""

CHANGELOG = """<ul class="linkList">
<li><a class="link" href="#20120704">20120704</a></li>
<li><a class="link" href="#20120703">20120703</a></li>
</ul>
<div><a name="20120704"></a></div>
<div class="changelog changelog1">
<div class="changelogTitle">20120704</div>
<div class="changelogDesc">Intro.</div>
<ul class="changelog">
<li>1
<ul class="changelog1">
<li>2
<ul class="changelog2">
<li>3</li>
<li>4</li>
</ul>
</li>
<li>5</li>
</ul>
</li>
</ul>
</div>
<div><a name="20120703"></a></div>
<div class="changelog changelog2">
<div class="changelogTitle">20120703</div>
<ul class="changelog">
<li>1
<ul class="changelog1">
<li>2
<ul class="changelog2">
<li>3</li>
</ul>
</li>
</ul>
</li>
</ul>
</div>
"""

GLOSSARY = """<ul class="linkList">
<li><a class="link" href="#term1">Term1</a></li>
<li><a class="link" href="#term2">Term2</a></li>
</ul>
<div><a name="term1"></a></div>
<div class="box box1">
<div class="boxTitle">Term1</div>
<div class="paragraph">
Definition 1.
</div>
</div>
<div><a name="term2"></a></div>
<div class="box box2">
<div class="boxTitle">Term2</div>
<div class="paragraph">
Definition 2. Paragraph 1.
</div>
<div class="paragraph">
Definition 2. Paragraph 2.
</div>
</div>
"""

GBUFFER = """<table class="gbuffer">
<tr>
<th>Address</th>
<th>Length</th>
<th>Purpose</th>
</tr>
<tr>
<td class="gbufAddress"><a name="24595" class="link" href="../asm/24595.html">24595</a></td>
<td class="gbufLength">1</td>
<td class="gbufDesc">
<div class="gbufDesc">Game status buffer entry</div>
</td>
</tr>
<tr>
<td class="gbufAddress"><a name="24596" class="link" href="../asm/24596.html">24596</a></td>
<td class="gbufLength">1</td>
<td class="gbufDesc">
<div class="gbufDesc">Another game status buffer entry</div>
</td>
</tr>
</table>
"""

REGISTERS_1 = """<table class="input">
<tr>
<td class="register">A</td>
<td class="registerContents">Some value</td>
</tr>
<tr>
<td class="register">B</td>
<td class="registerContents">Some other value</td>
</tr>
</table>
"""

REGISTERS_2 = """<table class="input">
<tr>
<th colspan="2">Input</th>
</tr>
<tr>
<td class="register">A</td>
<td class="registerContents">Some value</td>
</tr>
<tr>
<td class="register">B</td>
<td class="registerContents">Some other value</td>
</tr>
</table>
<table class="output">
<tr>
<th colspan="2">Output</th>
</tr>
<tr>
<td class="register">D</td>
<td class="registerContents">The result</td>
</tr>
<tr>
<td class="register">E</td>
<td class="registerContents">Result flags</td>
</tr>
</table>
"""

class TestHtmlWriter(HtmlWriter):
    def write_image(self, img_file, udgs=(), crop_rect=(), scale=2, mask=False):
        self.image_writer.write_image(udgs, scale, mask, *crop_rect)

class MockSkoolParser:
    def __init__(self, snapshot, entries=(), memory_map=()):
        self.snapshot = snapshot
        self.entries = entries
        self.memory_map = memory_map
        self.skoolfile = ''

class MockImageWriter:
    def __init__(self):
        self.build_images = True
        self.default_format = 'png'

    def write_image(self, udg_array, scale, mask, x, y, width, height):
        self.udg_array = udg_array
        self.scale = scale
        self.mask = mask
        self.x = x
        self.y = y
        self.width = width
        self.height = height

class HtmlWriterTest(SkoolKitTestCase):
    def tearDown(self):
        rmtree(ODIR, True)
        SkoolKitTestCase.tearDown(self)

    def read_file(self, fname, lines=False):
        f = open(join(ODIR, GAMEDIR, fname), 'r')
        if lines:
            contents = [line.rstrip('\n') for line in f]
        else:
            contents = f.read()
        f.close()
        return contents

    def assert_files_equal(self, d_fname, subs):
        d_html_lines = self.read_file(d_fname, True)
        t_html = subs['content']
        body_class = subs.get('body_class')
        body_class_attr = ' class="{0}"'.format(body_class) if body_class else ''
        subs['body'] = '<body{0}>'.format(body_class_attr)
        js = subs.get('js')
        subs['script'] = '\n<script type="text/javascript" src="{0}"></script>'.format(js) if js else ''
        subs.setdefault('header', subs['title'])
        t_html_lines = (HEADER + t_html + FOOTER).format(**subs).split('\n')
        self.assertEqual(d_html_lines, t_html_lines)

    def _test_reference_macro(self, macro, def_link_text, page):
        writer = self._get_writer()
        for link_text in ('', '(test)', '(test (nested) parentheses)'):
            for anchor in ('', '#test'):
                output = writer.expand('#{0}{1}{2}'.format(macro, anchor, link_text), ASMDIR)
                self.link_equals(output, '../{0}/{1}.html{2}'.format(REFERENCE_DIR, page, anchor), link_text[1:-1] or def_link_text)

    def _test_call(self, cwd, arg1, arg2):
        # Method used to test the #CALL macro
        return '{0}{1}'.format(cwd, arg2 * arg1)

    def _get_writer(self, use_ref=False, snapshot=None):
        ref_parser = RefParser()
        if use_ref:
            ref_parser.parse(TEST_REF)
        if snapshot is not None:
            skool_parser = MockSkoolParser(snapshot)
        else:
            skool_parser = SkoolParser(TEST_SKOOL, html=True)
        file_info = FileInfo(ODIR, GAMEDIR, False)
        return TestHtmlWriter(skool_parser, ref_parser, file_info, MockImageWriter())

    def _assert_scr_equal(self, game, x0=0, y0=0, w=32, h=24):
        snapshot = game.snapshot[:]
        scr = game.screenshot(x0, y0, w, h)
        self.assertEqual(len(scr),  min((h, 24 - y0)))
        self.assertEqual(len(scr[0]), min((w, 32 - x0)))
        for j, row in enumerate(scr):
            for i, udg in enumerate(row):
                x, y = x0 + i, y0 + j
                df_addr = 16384 + 2048 * (y // 8) + 32 * (y % 8) + x
                af_addr = 22528 + 32 * y + x
                self.assertEqual(udg.data, snapshot[df_addr:df_addr + 2048:256], 'Graphic data for cell at ({0},{1}) is incorrect'.format(x, y))
                self.assertEqual(udg.attr, snapshot[af_addr], 'Attribute byte for cell at ({0},{1}) is incorrect'.format(x, y))

    def link_equals(self, html, href, text):
        self.assertEqual(html, '<a class="link" href="{0}">{1}</a>'.format(href, text))

    def img_equals(self, html, alt, src):
        self.assertEqual(html, '<img alt="{0}" src="{1}" />'.format(alt, src))

    def _check_image(self, image_writer, udg_array, scale, mask, x, y, width, height):
        self.assertEqual(image_writer.scale, scale)
        self.assertEqual(image_writer.mask, mask)
        self.assertEqual(image_writer.x, x)
        self.assertEqual(image_writer.y, y)
        self.assertEqual(image_writer.width, width)
        self.assertEqual(image_writer.height, height)
        self.assertEqual(len(image_writer.udg_array), len(udg_array))
        for i, row in enumerate(image_writer.udg_array):
            exp_row = udg_array[i]
            self.assertEqual(len(row), len(exp_row))
            for j, udg in enumerate(row):
                exp_udg = exp_row[j]
                self.assertEqual(udg.attr, exp_udg.attr)
                self.assertEqual(udg.data, exp_udg.data)
                self.assertEqual(udg.mask, exp_udg.mask)

    def test_get_screenshot(self):
        snapshot = [0] * 65536
        for a in range(16384, 23296):
            snapshot[a] = random.randint(0, 255)
        writer = self._get_writer(snapshot=snapshot)
        self._assert_scr_equal(writer)
        self._assert_scr_equal(writer, 1, 2, 12, 10)
        self._assert_scr_equal(writer, 10, 10)

    def test_get_font_udg_array(self):
        snapshot = [0] * 65536
        char1 = [1, 2, 3, 4, 5, 6, 7, 8]
        char2 = [8, 7, 6, 5, 4, 3, 2, 1]
        chars = [char1, char2]
        char_data = []
        for char in chars:
            char_data.extend(char)
        address = 32768
        snapshot[address:address + sum(len(c) for c in chars)] = char_data
        writer = self._get_writer(snapshot=snapshot)
        attr = 56
        width, font_udg_array = writer.get_font_udg_array(address, len(chars), attr)
        self.assertEqual(width, 8 * len(chars))
        self.assertEqual(len(font_udg_array[0]), len(chars))
        for i, udg in enumerate(font_udg_array[0]):
            self.assertEqual(udg.attr, attr)
            self.assertEqual(udg.data, chars[i])

    def test_get_instruction_addr_str(self):
        writer = self._get_writer()
        self.assertEqual(writer.get_instruction_addr_str(24576), '24576')
        self.assertEqual(writer.get_instruction_addr_str(24579), '6003')

    def test_ref_parsing(self):
        writer = self._get_writer(use_ref=True)

        # [Info]
        self.assertEqual(writer.info['Release'], 'Test HTML disassembly')
        self.assertEqual(writer.info['Copyright'], 'Me, 2012')

        # [Links]
        self.assertEqual(writer.links['Bugs'], ('Bugs', ' (program errors)'))
        self.assertEqual(writer.links['Pokes'], ('Pokes [with square brackets in the link text]', ' (cheats)'))

        # [MemoryMap:*]
        self.assertTrue('TestMap' in writer.memory_map_names)
        self.assertTrue('TestMap' in writer.memory_maps)
        self.assertEqual(writer.memory_maps['TestMap'], {'EntryTypes': 'w', 'Name': 'TestMap'})

    def test_parse_image_params(self):
        writer = self._get_writer()
        def_crop_rect = (0, 0, None, None)

        text = '1,2,$3'
        end, img_path, crop_rect, p1, p2, p3 = writer.parse_image_params(text, 0, 3)
        self.assertEqual(img_path, None)
        self.assertEqual(crop_rect, def_crop_rect)
        self.assertEqual((p1, p2, p3), (1, 2, 3))
        self.assertEqual(end, len(text))

        text = '4,$a,6{0,1,24,32}'
        end, img_path, crop_rect, p1, p2, p3, p4, p5 = writer.parse_image_params(text, 0, 5, (7, 8))
        self.assertEqual(img_path, None)
        self.assertEqual(crop_rect, (0, 1, 24, 32))
        self.assertEqual((p1, p2, p3, p4, p5), (4, 10, 6, 7, 8))
        self.assertEqual(end, len(text))

        text = '$ff,8,9(img)'
        end, img_path, crop_rect, p1, p2, p3 = writer.parse_image_params(text, 0, 3)
        self.assertEqual(img_path, 'images/udgs/img.png')
        self.assertEqual(crop_rect, def_crop_rect)
        self.assertEqual((p1, p2, p3), (255, 8, 9))
        self.assertEqual(end, len(text))

        text = '0,1{1,2}(scr)'
        end, img_path, crop_rect, p1, p2 = writer.parse_image_params(text, 0, 2, path_id='ScreenshotImagePath')
        self.assertEqual(img_path, 'images/scr/scr.png')
        self.assertEqual(crop_rect, (1, 2, None, None))
        self.assertEqual((p1, p2), (0, 1))
        self.assertEqual(end, len(text))

        text = '0,1,2,3{1,2}(x)'
        end, img_path, crop_rect, p1, p2, p3 = writer.parse_image_params(text, 0, 3)
        self.assertEqual(img_path, None)
        self.assertEqual(crop_rect, def_crop_rect)
        self.assertEqual((p1, p2, p3), (0, 1, 2))
        self.assertEqual(end, 5)

    def test_image_path(self):
        writer = self._get_writer()
        def_img_format = writer.default_image_format

        img_path = writer.image_path('img')
        self.assertEqual(img_path, '{0}/img.{1}'.format(writer.paths['UDGImagePath'], def_img_format))

        path_id = 'ScreenshotImagePath'
        img_path = writer.image_path('img.gif', path_id)
        self.assertEqual(img_path, '{0}/img.gif'.format(writer.paths[path_id]))

        path_id = 'UnknownImagePath'
        fname = 'img.png'
        try:
            writer.image_path(fname, path_id)
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], "Unknown path ID '{0}' for image file '{1}'".format(path_id, fname))

    def test_flip(self):
        udg = Udg(0, [1, 2, 4, 8, 16, 32, 64, 128], [1, 2, 4, 8, 16, 32, 64, 128])
        udg.flip(0)
        self.assertEqual(udg.data, [1, 2, 4, 8, 16, 32, 64, 128])
        self.assertEqual(udg.mask, [1, 2, 4, 8, 16, 32, 64, 128])

        udg = Udg(0, [1, 2, 4, 8, 16, 32, 64, 128], [1, 2, 4, 8, 16, 32, 64, 128])
        udg.flip(1)
        self.assertEqual(udg.data, [128, 64, 32, 16, 8, 4, 2, 1])
        self.assertEqual(udg.mask, [128, 64, 32, 16, 8, 4, 2, 1])

        udg = Udg(0, [1, 2, 3, 4, 5, 6, 7, 8], [2, 4, 6, 8, 10, 12, 14, 16])
        udg.flip(2)
        self.assertEqual(udg.data, [8, 7, 6, 5, 4, 3, 2, 1])
        self.assertEqual(udg.mask, [16, 14, 12, 10, 8, 6, 4, 2])

        udg = Udg(0, [1, 2, 3, 4, 5, 6, 7, 8], [8, 7, 6, 5, 4, 3, 2, 1])
        udg.flip(3)
        self.assertEqual(udg.data, [16, 224, 96, 160, 32, 192, 64, 128])
        self.assertEqual(udg.mask, [128, 64, 192, 32, 160, 96, 224, 16])

    def test_flip_udgs(self):
        writer = self._get_writer()
        udg1 = Udg(0, [1, 2, 4, 8, 16, 32, 64, 128], [1, 2, 4, 8, 16, 32, 64, 128])
        udg2 = Udg(0, [1, 2, 3, 4, 5, 6, 7, 8], [2, 4, 6, 8, 10, 12, 14, 16])
        udg3 = Udg(0, [1, 2, 3, 4, 5, 6, 7, 8], [8, 7, 6, 5, 4, 3, 2, 1])
        udg4 = Udg(0, [8, 7, 6, 5, 4, 3, 2, 1], [255, 254, 253, 252, 251, 250, 249, 248])

        udgs = [[udg1.copy(), udg2.copy()], [udg3.copy(), udg4.copy()]]
        writer.flip_udgs(udgs, 0)
        self.assertEqual(udgs, [[udg1, udg2], [udg3, udg4]])

        udgs = [[udg1.copy(), udg2.copy()], [udg3.copy(), udg4.copy()]]
        writer.flip_udgs(udgs, 1)
        udg1_f, udg2_f, udg3_f, udg4_f = udg1.copy(), udg2.copy(), udg3.copy(), udg4.copy()
        udg1_f.flip(1)
        udg2_f.flip(1)
        udg3_f.flip(1)
        udg4_f.flip(1)
        self.assertEqual(udgs, [[udg2_f, udg1_f], [udg4_f, udg3_f]])

        udgs = [[udg1.copy(), udg2.copy()], [udg3.copy(), udg4.copy()]]
        writer.flip_udgs(udgs, 2)
        udg1_f, udg2_f, udg3_f, udg4_f = udg1.copy(), udg2.copy(), udg3.copy(), udg4.copy()
        udg1_f.flip(2)
        udg2_f.flip(2)
        udg3_f.flip(2)
        udg4_f.flip(2)
        self.assertEqual(udgs, [[udg3_f, udg4_f], [udg1_f, udg2_f]])

        udgs = [[udg1.copy(), udg2.copy()], [udg3.copy(), udg4.copy()]]
        writer.flip_udgs(udgs, 3)
        udg1_f, udg2_f, udg3_f, udg4_f = udg1.copy(), udg2.copy(), udg3.copy(), udg4.copy()
        udg1_f.flip(3)
        udg2_f.flip(3)
        udg3_f.flip(3)
        udg4_f.flip(3)
        self.assertEqual(udgs, [[udg4_f, udg3_f], [udg2_f, udg1_f]])

    def test_rotate(self):
        udg = Udg(0, [1, 2, 4, 8, 16, 32, 64, 128], [1, 2, 4, 8, 16, 32, 64, 128])
        udg.rotate(0)
        self.assertEqual(udg.data, [1, 2, 4, 8, 16, 32, 64, 128])
        self.assertEqual(udg.mask, [1, 2, 4, 8, 16, 32, 64, 128])

        udg = Udg(0, [1, 2, 4, 8, 16, 32, 64, 128], [1, 2, 4, 8, 16, 32, 64, 128])
        udg.rotate(1)
        self.assertEqual(udg.data, [128, 64, 32, 16, 8, 4, 2, 1])
        self.assertEqual(udg.mask, [128, 64, 32, 16, 8, 4, 2, 1])

        udg = Udg(0, [1, 2, 3, 4, 5, 6, 7, 8], [8, 7, 6, 5, 4, 3, 2, 1])
        udg.rotate(2)
        self.assertEqual(udg.data, [16, 224, 96, 160, 32, 192, 64, 128])
        self.assertEqual(udg.mask, [128, 64, 192, 32, 160, 96, 224, 16])

        udg = Udg(0, [1, 2, 3, 4, 5, 6, 7, 8], [255, 254, 253, 252, 251, 250, 249, 248])
        udg.rotate(3)
        self.assertEqual(udg.data, [170, 102, 30, 1, 0, 0, 0, 0])
        self.assertEqual(udg.mask, [170, 204, 240, 255, 255, 255, 255, 255])

    def test_rotate_udgs(self):
        writer = self._get_writer()
        udg1 = Udg(0, [1, 2, 4, 8, 16, 32, 64, 128], [1, 2, 4, 8, 16, 32, 64, 128])
        udg2 = Udg(0, [1, 2, 3, 4, 5, 6, 7, 8], [2, 4, 6, 8, 10, 12, 14, 16])
        udg3 = Udg(0, [1, 2, 3, 4, 5, 6, 7, 8], [8, 7, 6, 5, 4, 3, 2, 1])
        udg4 = Udg(0, [8, 7, 6, 5, 4, 3, 2, 1], [255, 254, 253, 252, 251, 250, 249, 248])

        udgs = [[udg1.copy(), udg2.copy()], [udg3.copy(), udg4.copy()]]
        writer.rotate_udgs(udgs, 0)
        self.assertEqual(udgs, [[udg1, udg2], [udg3, udg4]])

        udgs = [[udg1.copy(), udg2.copy()], [udg3.copy(), udg4.copy()]]
        writer.rotate_udgs(udgs, 1)
        udg1_r, udg2_r, udg3_r, udg4_r = udg1.copy(), udg2.copy(), udg3.copy(), udg4.copy()
        udg1_r.rotate(1)
        udg2_r.rotate(1)
        udg3_r.rotate(1)
        udg4_r.rotate(1)
        self.assertEqual(udgs, [[udg3_r, udg1_r], [udg4_r, udg2_r]])

        udgs = [[udg1.copy(), udg2.copy()], [udg3.copy(), udg4.copy()]]
        writer.rotate_udgs(udgs, 2)
        udg1_r, udg2_r, udg3_r, udg4_r = udg1.copy(), udg2.copy(), udg3.copy(), udg4.copy()
        udg1_r.rotate(2)
        udg2_r.rotate(2)
        udg3_r.rotate(2)
        udg4_r.rotate(2)
        self.assertEqual(udgs, [[udg4_r, udg3_r], [udg2_r, udg1_r]])

        udgs = [[udg1.copy(), udg2.copy()], [udg3.copy(), udg4.copy()]]
        writer.rotate_udgs(udgs, 3)
        udg1_r, udg2_r, udg3_r, udg4_r = udg1.copy(), udg2.copy(), udg3.copy(), udg4.copy()
        udg1_r.rotate(3)
        udg2_r.rotate(3)
        udg3_r.rotate(3)
        udg4_r.rotate(3)
        self.assertEqual(udgs, [[udg2_r, udg4_r], [udg1_r, udg3_r]])

    def test_macro_bug(self):
        self._test_reference_macro('BUG', 'bug', 'bugs')

    def test_macro_call(self):
        writer = self._get_writer()
        writer.test_call = self._test_call
        output = writer.expand('#CALL:test_call(10,test)', ASMDIR)
        self.assertEqual(output, self._test_call(ASMDIR, 10, 'test'))

    def test_macro_chr(self):
        writer = self._get_writer()

        output = writer.expand('#CHR169', ASMDIR)
        self.assertEqual(output, '&#169;')

        output = writer.expand('#CHR(163)1984', ASMDIR)
        self.assertEqual(output, '&#163;1984')

    def test_macro_d(self):
        writer = self._get_writer()
        output = writer.expand('#D24592', ASMDIR)
        self.assertEqual(output, 'The final routine')

    def test_macro_erefs(self):
        writer = self._get_writer()
        output = writer.expand('#EREFS24582', ASMDIR)
        self.assertEqual(output, 'routines at <a class="link" href="24586.html">24586</a> and <a class="link" href="24592.html">24592</a>')

    def test_macro_fact(self):
        self._test_reference_macro('FACT', 'fact', 'facts')

    def test_macro_font(self):
        snapshot = [0] * 65536
        writer = self._get_writer(snapshot=snapshot)

        output = writer.expand('#FONT32768,96', ASMDIR)
        self.img_equals(output, 'font', '../images/font/font.png')

        img_fname = 'font2'
        output = writer.expand('#FONT55584,96({0})'.format(img_fname), ASMDIR)
        self.img_equals(output, img_fname, '../images/font/{0}.png'.format(img_fname))

        img_fname = 'font3'
        font_addr = 32768
        char1 = [1, 2, 3, 4, 5, 6, 7, 8]
        char2 = [8, 7, 6, 5, 4, 3, 2, 1]
        char3 = [1, 3, 7, 15, 31, 63, 127, 255]
        chars = (char1, char2, char3)
        for i, char in enumerate(chars):
            snapshot[font_addr + i * 8:font_addr + (i + 1) * 8] = char
        attr = 4
        scale = 2
        x, y, w, h = 1, 2, 3, 4
        macro = '#FONT{0},{1},{2},{3}{{{4},{5},{6},{7}}}({8})'.format(font_addr, len(chars), attr, scale, x, y, w, h, img_fname)
        output = writer.expand(macro, ASMDIR)
        self.img_equals(output, img_fname, '../images/font/{0}.png'.format(img_fname))
        udg_array = [[Udg(4, char) for char in chars]]
        self._check_image(writer.image_writer, udg_array, scale, False, x, y, w, h)

    def test_macro_html(self):
        writer = self._get_writer()

        delimiters = {
            '(': ')',
            '[': ']',
            '{': '}'
        }
        for text in('', 'See <a href="url">this</a>', 'A &gt; B'):
            for delim1 in '([{!@$%^*_-+|':
                delim2 = delimiters.get(delim1, delim1)
                output = writer.expand('#HTML{0}{1}{2}'.format(delim1, text, delim2), ASMDIR)
                self.assertEqual(output, text)

        output = writer.expand('#HTML?#CHR169?', ASMDIR)
        self.assertEqual(output, '&#169;')

    def test_macro_link(self):
        writer = self._get_writer(True)

        link_text = 'bugs'
        output = writer.expand('#LINK:Bugs({0})'.format(link_text), ASMDIR)
        self.link_equals(output, '../{0}/bugs.html'.format(REFERENCE_DIR), link_text)

        link_text = 'pokes'
        anchor = '#poke1'
        output = writer.expand('#LINK:Pokes{0}({1})'.format(anchor, link_text), ASMDIR)
        self.link_equals(output, '../{0}/pokes.html{1}'.format(REFERENCE_DIR, anchor), link_text)

        output = writer.expand('#LINK:CustomPage1()', ASMDIR)
        self.link_equals(output, '../page.html', 'Custom page')

        output = writer.expand('#LINK:CustomPage2()', ASMDIR)
        self.link_equals(output, '../page2.html', 'Custom page 2')

    def test_macro_name(self):
        writer = self._get_writer()
        anchor = 'testing'
        output = writer.expand('#NAME{0}'.format(anchor), ASMDIR)
        self.assertEqual(output, '<a name="{0}"></a>'.format(anchor))

    def test_macro_poke(self):
        self._test_reference_macro('POKE', 'poke', 'pokes')

    def test_macro_pokes(self):
        writer = self._get_writer()
        snapshot = writer.snapshot

        output = writer.expand('#POKES32768,255', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768], 255)

        output = writer.expand('#POKES32768,254,10', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768:32778], [254] * 10)

        output = writer.expand('#POKES32768,253,20,2', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[32768:32808:2], [253] * 20)

        output = writer.expand('#POKES49152,1;49153,2', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(snapshot[49152:49154], [1, 2])

    def test_macro_pops(self):
        writer = self._get_writer()
        addr, byte = 49152, 128
        writer.snapshot[addr] = byte
        writer.push_snapshot('test')
        writer.snapshot[addr] = (byte + 127) % 256
        output = writer.expand('#POPS', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(writer.snapshot[addr], byte)

    def test_macro_pushs(self):
        writer = self._get_writer()
        addr, byte = 32768, 64
        writer.snapshot[addr] = byte
        output = writer.expand('#PUSHStest', ASMDIR)
        self.assertEqual(output, '')
        self.assertEqual(writer.snapshot[addr], byte)
        writer.snapshot[addr] = (byte + 127) % 256
        writer.pop_snapshot()
        self.assertEqual(writer.snapshot[addr], byte)

    def test_macro_r(self):
        writer = self._get_writer(True)

        output = writer.expand('#R24576', ASMDIR)
        self.link_equals(output, '24576.html', '24576')

        link_text = 'Testing1'
        output = writer.expand('#R24576({0})'.format(link_text), ASMDIR)
        self.link_equals(output, '24576.html', link_text)

        output = writer.expand('#R24576', 'other')
        self.link_equals(output, '../{0}/24576.html'.format(ASMDIR), '24576')

        output = writer.expand('#R24579', ASMDIR)
        self.link_equals(output, '24579.html', '6003')

        output = writer.expand('#R24580', ASMDIR)
        self.link_equals(output, '24579.html#24580', '6004')

        output = writer.expand('#R32768@other', ASMDIR)
        self.link_equals(output, '../other/32768.html', '32768')

        link_text = 'Testing2'
        anchor = 'testing3'
        output = writer.expand('#R32768@other#{0}({1})'.format(anchor, link_text), ASMDIR)
        self.link_equals(output, '../other/32768.html#{0}'.format(anchor), link_text)

    def test_macro_refs(self):
        writer = self._get_writer()
        output = writer.expand('#REFS24581', ASMDIR)
        self.assertEqual(output, 'routines at <a class="link" href="24583.html">24583</a>, <a class="link" href="24586.html">24586</a> and <a class="link" href="24592.html">24592</a>')

    def test_macro_reg(self):
        writer = self._get_writer()
        for reg in ("a", "b", "c", "d", "e", "h", "l", "i", "r", "ixl", "ixh", "iyl", "iyh", "b'", "c'", "d'", "e'", "h'", "l'", "bc", "de", "hl", "sp", "ix", "iy", "bc'", "de'", "hl'"):
            output = writer.expand('#REG{0}'.format(reg), ASMDIR)
            self.assertEqual(output, '<span class="register">{0}</span>'.format(reg.upper()))

    def test_macro_scr(self):
        snapshot = [0] * 65536
        writer = self._get_writer(snapshot=snapshot)

        output = writer.expand('#SCR', ASMDIR)
        self.img_equals(output, 'scr', '../images/scr/scr.png')

        scr_fname = 'scr2'
        output = writer.expand('#SCR2,0,0,10,10({0})'.format(scr_fname), ASMDIR)
        self.img_equals(output, scr_fname, '../images/scr/{0}.png'.format(scr_fname))

        scr_fname = 'scr3'
        data = [128, 64, 32, 16, 8, 4, 2, 1]
        attr = 48
        snapshot[16384:18432:256] = data
        snapshot[22528] = attr
        scale = 2
        x, y, w, h = 1, 2, 5, 6
        macro = '#SCR{0},0,0,1,1{{{1},{2},{3},{4}}}({5})'.format(scale, x, y, w, h, scr_fname)
        output = writer.expand(macro, ASMDIR)
        self.img_equals(output, scr_fname, '../images/scr/{0}.png'.format(scr_fname))
        udg_array = [[Udg(attr, data)]]
        self._check_image(writer.image_writer, udg_array, scale, False, x, y, w, h)

    def test_macro_space(self):
        writer = self._get_writer()

        output = writer.expand('#SPACE', ASMDIR)
        self.assertEqual(output, '&#160;')

        num = 10
        output = writer.expand('#SPACE{0}'.format(num), ASMDIR)
        self.assertEqual(output, '&#160;' * num)

        num = 7
        output = writer.expand('1#SPACE({0})1'.format(num), ASMDIR)
        self.assertEqual(output, '1{0}1'.format('&#160;' * num))

    def test_macro_table(self):
        writer = self._get_writer()
        output = writer.expand('#TABLE{0}\nTABLE#'.format(TABLE_SRC), ASMDIR)
        self.assertEqual(output, TABLE_HTML)

    def test_macro_udg(self):
        snapshot = [0] * 65536
        writer = self._get_writer(snapshot=snapshot)

        udg_fname = 'udg32768_56x4'
        output = writer.expand('#UDG32768', ASMDIR)
        self.img_equals(output, udg_fname, '../{0}/{1}.png'.format(UDGDIR, udg_fname))

        udg_fname = 'udg40000_2x3'
        output = writer.expand('#UDG40000,2,3', ASMDIR)
        self.img_equals(output, udg_fname, '../{0}/{1}.png'.format(UDGDIR, udg_fname))

        udg_fname = 'test_udg'
        output = writer.expand('#UDG32768,2,6,1,0:49152,2({0})'.format(udg_fname), ASMDIR)
        self.img_equals(output, udg_fname, '../{0}/{1}.png'.format(UDGDIR, udg_fname))

        udg_fname = 'test_udg2'
        udg_addr = 32768
        attr = 2
        scale = 1
        step = 1
        inc = 0
        mask_addr = 32776
        x, y, w, h = 2, 1, 3, 4
        udg_data = [136] * 8
        udg_mask = [255] * 8
        snapshot[udg_addr:udg_addr + 8 * step:step] = udg_data
        snapshot[mask_addr:mask_addr + 8 * step:step] = udg_mask
        macro = '#UDG{0},{1},{2},{3},{4}:{5},{6}{{{7},{8},{9},{10}}}({11})'.format(udg_addr, attr, scale, step, inc, mask_addr, step, x, y, w, h, udg_fname)
        output = writer.expand(macro, ASMDIR)
        self.img_equals(output, udg_fname, '../{0}/{1}.png'.format(UDGDIR, udg_fname))
        udg_array = [[Udg(attr, udg_data, udg_mask)]]
        self._check_image(writer.image_writer, udg_array, scale, True, x, y, w, h)

    def test_macro_udgarray(self):
        snapshot = [0] * 65536
        writer = self._get_writer(snapshot=snapshot)

        udg_fname = 'test_udg_array'
        output = writer.expand('#UDGARRAY8;32768-32784-1-8({0})'.format(udg_fname), ASMDIR)
        self.img_equals(output, udg_fname, '../{0}/{1}.png'.format(UDGDIR, udg_fname))

        udg_fname = 'test_udg_array2'
        output = writer.expand('#UDGARRAY8,56,2,256,0;32768;32769;32770;32771;32772x12({0})'.format(udg_fname), ASMDIR)
        self.img_equals(output, udg_fname, '../{0}/{1}.png'.format(UDGDIR, udg_fname))

        udg_fname = 'test_udg_array3'
        udg_addr = 32768
        mask_addr = 32769
        width = 2
        attr = 5
        scale = 1
        step = 256
        inc = 0
        x, y, w, h = 4, 6, 8, 5
        udg_data = [195] * 8
        udg_mask = [255] * 8
        snapshot[udg_addr:udg_addr + 8 * step:step] = udg_data
        snapshot[mask_addr:mask_addr + 8 * step:step] = udg_mask
        macro = '#UDGARRAY{0},{1},{2},{3},{4};{5}x4:{6}x4{{{7},{8},{9},{10}}}({11})'.format(width, attr, scale, step, inc, udg_addr, mask_addr, x, y, w, h, udg_fname)
        output = writer.expand(macro, ASMDIR)
        self.img_equals(output, udg_fname, '../{0}/{1}.png'.format(UDGDIR, udg_fname))
        udg_array = [[Udg(attr, udg_data, udg_mask)] * width] * 2
        self._check_image(writer.image_writer, udg_array, scale, True, x, y, w, h)

    def test_macro_udgtable(self):
        writer = self._get_writer()
        output = writer.expand('#UDGTABLE{0}\nUDGTABLE#'.format(TABLE_SRC), ASMDIR)
        self.assertEqual(output, TABLE_HTML)

    def test_html_escape(self):
        # Check that HTML characters from the skool file are escaped
        writer = self._get_writer()
        fname = 'test.html'
        writer.write_entry(ASMDIR, writer.entries[24576], fname)
        html = self.read_file(join(ASMDIR, fname))
        self.assertTrue('a &lt;= b and b &gt;= c' in html)

    def test_html_no_escape(self):
        # Check that HTML characters from the ref file are not escaped
        writer = self._get_writer(True)
        writer.write_bugs()
        html = self.read_file(join(REFERENCE_DIR, 'bugs.html'))
        self.assertTrue('<p>Hello</p>' in html)

    def test_write_changelog(self):
        writer = self._get_writer(True)
        writer.write_changelog()
        subs = {
            'title': 'Changelog',
            'path': '../',
            'body_class': 'changelog',
            'content': CHANGELOG
        }
        self.assert_files_equal(join(REFERENCE_DIR, 'changelog.html'), subs)

    def test_write_glossary(self):
        writer = self._get_writer(True)
        writer.write_glossary()
        subs = {
            'title': 'Glossary',
            'path': '../',
            'body_class': 'glossary',
            'content': GLOSSARY
        }
        self.assert_files_equal(join(REFERENCE_DIR, 'glossary.html'), subs)

    def test_write_gbuffer(self):
        writer = self._get_writer(True)
        writer.write_gbuffer()
        subs = {
            'title': 'Game status buffer',
            'path': '../',
            'body_class': 'gbuffer',
            'content': GBUFFER
        }
        self.assert_files_equal(join(BUFFERS_DIR, 'gbuffer.html'), subs)

    def test_write_graphics(self):
        writer = self._get_writer(True)
        writer.write_graphics()
        subs = {
            'title': 'Graphics',
            'path': '../',
            'body_class': 'graphics',
            'content': '<em>This is the graphics page.</em>\n'
        }
        self.assert_files_equal(join(GRAPHICS_DIR, 'graphics.html'), subs)

    def test_write_page(self):
        writer = self._get_writer(True)
        writer.write_page('CustomPage1')
        subs = {
            'title': 'Custom page',
            'path': '',
            'js': 'test-html.js',
            'content': '<b>This is the content of custom page 1.</b>\n'
        }
        self.assert_files_equal('page.html', subs)

    def test_get_udg_addresses(self):
        writer = self._get_writer(snapshot=())
        addr_specs = [
            (0, 1, [0]),
            ('1', 1, [1]),
            ('2x3', 1, [2] * 3),
            ('0-3', 1, [0, 1, 2, 3]),
            ('0-2x3', 1, [0, 1, 2] * 3),
            ('0-6-2', 1, [0, 2, 4, 6]),
            ('0-6-3x2', 1, [0, 3, 6] * 2),
            ('0-49-1-16', 2, [0, 1, 16, 17, 32, 33, 48, 49]),
            ('0-528-8-256x4', 3, [0, 8, 16, 256, 264, 272, 512, 520, 528] * 4)
        ]
        for addr_spec, width, exp_addresses in addr_specs:
            self.assertEqual(writer._get_udg_addresses(addr_spec, width), exp_addresses)

    def test_write_registers(self):
        writer = self._get_writer(snapshot=())

        # Traditional
        registers = []
        registers.append(Register('', 'A', 'Some value'))
        registers.append(Register('', 'B', 'Some other value'))
        stream = StringIO()
        writer.write_registers(stream, registers, ASMDIR)
        self.assertEqual(stream.getvalue(), REGISTERS_1)

        # With prefixes
        writer.game_vars['InputRegisterTableHeader'] = 'Input'
        writer.game_vars['OutputRegisterTableHeader'] = 'Output'
        registers = []
        registers.append(Register('Input', 'A', 'Some value'))
        registers.append(Register('', 'B', 'Some other value'))
        registers.append(Register('Output', 'D', 'The result'))
        registers.append(Register('', 'E', 'Result flags'))
        stream = StringIO()
        writer.write_registers(stream, registers, ASMDIR)
        self.assertEqual(stream.getvalue(), REGISTERS_2)

if __name__ == '__main__':
    unittest.main()
