# -*- coding: utf-8 -*-
import os
import unittest

from skoolkittest import SkoolKitTestCase
from skoolkit import SkoolKitError

TEST_BIN = 'test-sna.bin'

TEST_CTL_G = """c 65472
c 65476
c 65478
c 65481
c 65490
c 65498
z 65505
c 65515
z 65517"""

TEST_CTL_G_HEX = """c $FFC0
c $FFC4
c $FFC6
c $FFC9
c $FFD2
c $FFDA
z $FFE1
c $FFEB
z $FFED"""

USAGE = """Usage: sna2skool.py [options] file

  Convert a binary (raw memory) file or a 48K SNA or Z80 snapshot into a skool
  file.

Options:
  -c FILE  Use FILE as the control file (default is file.ctl)
  -T FILE  Use FILE as the skool file template (default is file.sft)
  -g FILE  Generate a control file in FILE
  -h       Write hexadecimal addresses in the generated control file
  -H       Write hexadecimal addresses and operands in the disassembly
  -L       Write the disassembly in lower case
  -s ADDR  Specify the address at which to start disassembling (default=16384)
  -o ADDR  Specify the origin address of file.bin (default: 65536 - length)
  -t       Show ASCII text in the comment fields
  -r       Don't add comments that list entry point referrers
  -n N     Set the max number of bytes per DEFB statement to N (default=8)
  -m M     Group DEFB blocks by addresses that divide M
  -z       Write bytes with leading zeroes in DEFB statements
  -l L     Set the max number of characters per DEFM statement to L (default=66)"""

class OptionsTest(SkoolKitTestCase):
    def _write_bin(self, data, ctls=None):
        binfile = self.write_bin_file(data, suffix='.bin')
        if ctls is not None:
            self._write_ctl(ctls, binfile[:-4] + '.ctl')
        return binfile

    def _write_ctl(self, ctls, ctlfile=None):
        return self.write_text_file('\n'.join(ctls), ctlfile)

    def test_invalid_option(self):
        output, error = self.run_sna2skool('-x dummy.bin', err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'))

    def test_no_arguments(self):
        output, error = self.run_sna2skool(err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'))

    def test_no_options(self):
        binfile = self._write_bin([201])
        lines = self._write_skool(binfile, 10)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org=65535')
        self.assertEqual(lines[2], '; Routine at 65535')
        self.assertEqual(lines[3][:10], 'c65535 RET')

    def test_nonexistent_files(self):
        error_tp = '{0}: file not found'

        nonexistent_bin = 'nonexistent.bin'
        try:
            self.run_sna2skool(nonexistent_bin)
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], error_tp.format(nonexistent_bin))

        nonexistent_ctl = 'nonexistent.ctl'
        try:
            self.run_sna2skool('-c {0}'.format(nonexistent_ctl))
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], error_tp.format(nonexistent_ctl))

        nonexistent_sft = 'nonexistent.sft'
        try:
            self.run_sna2skool('-T {0}'.format(nonexistent_sft))
            self.fail()
        except SkoolKitError as e:
            self.assertEqual(e.args[0], error_tp.format(nonexistent_sft))

    def test_option_H(self):
        data = [62, 254]           # $FFFB LD A,$FE
        data.extend((195, 1, 128)) # $FFFD JP $8001
        binfile = self._write_bin(data, ['c $FFFB'])
        lines = self._write_skool('-H {0}'.format(binfile), 5)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org=$FFFB')
        self.assertEqual(lines[3][:15], 'c$FFFB LD A,$FE')
        self.assertEqual(lines[4][:15], ' $FFFD JP $8001')

    def test_option_L(self):
        data = [0]      # 65534 nop
        data.append(65) # 65535 defm "A"
        binfile = self._write_bin(data, ['c 65534', 'T 65535'])
        lines = self._write_skool('-L {0}'.format(binfile), 5)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org=65534')
        self.assertEqual(lines[3][:10], 'c65534 nop')
        self.assertEqual(lines[4][:15], ' 65535 defm "A"')

    def test_option_T(self):
        data = [0] * 6
        data.append(120)           # 65532 LD A,B
        data.extend((195, 6, 128)) # 65533 JP 32774
        binfile = self._write_bin(data)
        sft = ['; Do stuff']
        sft.append('cC65532,1;20 Blah')
        sft.append(' C65533,3;20 Yah')
        sftfile = self.write_text_file('\n'.join(sft))
        lines = self._write_skool('-T {0} {1}'.format(sftfile, binfile), 2)
        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0], sft[0])
        self.assertEqual(lines[1], 'c65532 LD A,B       ; Blah')
        self.assertEqual(lines[2], ' 65533 JP 32774     ; Yah')

    def test_option_c(self):
        org = 65530
        binfile = self._write_bin([0] * 10)
        ctlfile = self._write_ctl(['c {0}'.format(org), 'i {0}'.format(org + 4)])
        lines = self._write_skool('-c {0} {1}'.format(ctlfile, binfile), 10)
        self.assertEqual(self._get_first_address(lines), str(org))

    def test_option_g(self):
        ctlfile = self.write_text_file()
        self._write_skool('-g {0} {1}'.format(ctlfile, TEST_BIN))
        with open(ctlfile, 'r') as f:
            lines = [line.rstrip() for line in f]
        self.assert_output_equal(TEST_CTL_G.split('\n'), lines)

    def test_option_h(self):
        ctlfile = self.write_text_file()
        self._write_skool('-g {0} -h {1}'.format(ctlfile, TEST_BIN))
        with open(ctlfile, 'r') as f:
            lines = [line.rstrip() for line in f]
        self.assert_output_equal(TEST_CTL_G_HEX.split('\n'), lines)

    def test_option_l(self):
        bin_len = 16
        binfile = self._write_bin([65] * bin_len, ['t {0}'.format(65536 - bin_len)])
        defm_len = 5
        defm = 'DEFM "{0}"'.format('A' * defm_len)
        lines = self._write_skool('-l {0} {1}'.format(defm_len, binfile), 6)
        for line in lines[3:6]:
            self.assertEqual(line[7:], defm)

    def test_option_m(self):
        bin_len = 25
        org = 65536 - bin_len
        binfile = self._write_bin([0] * bin_len, ['b {0}'.format(org)])
        defb_mod = 4
        first_defb_len = 4 - (org % defb_mod)
        first_defb = 'DEFB {0}'.format(','.join([str(b) for b in [0] * first_defb_len]))
        defb_len = 8
        defb = 'DEFB {0}'.format(','.join([str(b) for b in [0] * defb_len]))
        lines = self._write_skool('-m {0} {1}'.format(defb_mod, binfile), 6)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org={0}'.format(org))
        self.assertEqual(lines[3], 'b{0} {1}'.format(org, first_defb))
        self.assertEqual(lines[4], ' {0} {1}'.format(org + first_defb_len, defb))
        self.assertEqual(lines[5], ' {0} {1}'.format(org + first_defb_len + defb_len, defb))

    def test_option_n(self):
        bin_len = 10
        org = 65536 - bin_len
        binfile = self._write_bin([0] * bin_len, ['b {0}'.format(org)])
        defb_len = 3
        defb = 'DEFB {0}'.format(','.join([str(b) for b in [0] * defb_len]))
        lines = self._write_skool('-n {0} {1}'.format(defb_len, binfile), 6)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org={0}'.format(org))
        for line in lines[3:6]:
            self.assertEqual(line[7:], defb)

    def test_option_o(self):
        data = [33, 0, 0, 201]
        binfile = self._write_bin(data)
        org = str(65536 - len(data) - 100)
        lines = self._write_skool('-o {0} {1}'.format(org, binfile), 4)
        self.assertEqual(self._get_first_address(lines), org)

    def test_option_r(self):
        data = [201]           # 65533 RET
        data.extend((24, 253)) # 65534 JR 65533
        binfile = self._write_bin(data, ['c 65533', 'c 65534'])
        lines = self._write_skool('-r {0}'.format(binfile), 4)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org=65533')
        self.assertEqual(lines[2], '; Routine at 65533')
        self.assertEqual(lines[3][:10], 'c65533 RET')

    def test_option_s(self):
        data = [0] * 4
        binfile = self._write_bin(data)
        start = '65534'
        lines = self._write_skool('-s {0} {1}'.format(start, binfile), 10)
        self.assertTrue(self._get_first_address(lines) == start)

    def test_option_t(self):
        binfile = self._write_bin([49, 54, 50]) # DEFM "162"
        skool = self._write_skool('-t {0}'.format(binfile), 10)
        self.assertTrue('c65533 LD SP,12854   ; [162]' in skool)

    def test_option_z(self):
        data = [0, 1, 2, 3, 4, 5]
        org = 65536 - len(data)
        binfile = self._write_bin(data, ['b {0}'.format(org)])
        defb = 'DEFB {0}'.format(','.join(['%03i' % b for b in data]))
        lines = self._write_skool('-z {0}'.format(binfile), 4)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org={0}'.format(org))
        self.assertEqual(lines[3], 'b{0} {1}'.format(org, defb))

    def test_sna(self):
        data = [0] * 27 # header
        data += [1, 2, 3] # 16384 LD BC,770
        data += [0] * 49149
        snafile = self.write_bin_file(data, suffix='.sna')
        self._write_ctl(['c 16384', 'i 16387'], snafile[:-4] + '.ctl')
        lines = self._write_skool(snafile, 4)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org=16384')
        self.assertEqual(lines[3][:16], 'c16384 LD BC,770')

    def _write_skool(self, args, split=0):
        output, error = self.run_sna2skool(args, out_lines=False)
        if error:
            self.fail(error)
        if split:
            return [line.rstrip() for line in output.split('\n', split)]
        return output

    def _get_first_address(self, lines):
        for line in lines:
            if line.startswith('c'):
                return line[1:6]

if __name__ == '__main__':
    unittest.main()
