# -*- coding: utf-8 -*-
import os
import unittest

from skoolkittest import SkoolKitTestCase
from skoolkit.snapshot import get_snapshot, SnapshotError

class SnapshotTest(SkoolKitTestCase):
    def _get_ram_block(self, data, compress, page=None):
        if compress:
            block = []
            prev_b = data[0]
            count = 1
            for b in data[1:] + [-1]:
                if b == prev_b:
                    if count < 255:
                        count += 1
                        continue
                if count > 4 or (prev_b == 237 and count > 1):
                    block += [237, 237, count, prev_b]
                else:
                    block += [prev_b] * count
                prev_b = b
                count = 1
        else:
            block = data
        if page is not None:
            length = len(block) if compress else 65535
            return [length % 256, length // 256, page] + block
        return block + [0, 237, 237, 0]

    def _test_z80(self, exp_ram, version, compress):
        if version == 1:
            header = [0] * 30
            header[6] = 255 # Set PC > 0 to indicate a v1 Z80 snapshot
            if compress:
                header[12] |= 32 # Signal that the RAM data block is compressed
            z80 = header + self._get_ram_block(exp_ram, compress)
        else:
            if version == 2:
                header = [0] * 55
                header[30] = 23 # Indicate a v2 Z80 snapshot
            else:
                header = [0] * 86
                header[30] = 54 # Indicate a v3 Z80 snapshot
            z80 = header + self._get_ram_block(exp_ram[:16384], compress, 8)
            z80 += self._get_ram_block(exp_ram[16384:32768], compress, 4)
            z80 += self._get_ram_block(exp_ram[32768:49152], compress, 5)
        tmp_z80 = self.write_bin_file(z80, suffix='.z80')
        snapshot = get_snapshot(tmp_z80)
        ram = snapshot[16384:]
        self.assertEqual(len(ram), 49152)
        self.assertEqual(ram, exp_ram)

    def test_z80v1(self):
        exp_ram = [n & 255 for n in range(49152)]
        self._test_z80(exp_ram, 1, False)

    def test_decompress_z80v1(self):
        exp_ram = [0] * 261 + [1] * 10 + [2] * 4 + [237, 4] + [237] * 2
        exp_ram += [128] * (49152 - len(exp_ram))
        self._test_z80(exp_ram, 1, True)

    def test_z80v2(self):
        exp_ram = [(n + 1) & 255 for n in range(49152)]
        self._test_z80(exp_ram, 2, False)

    def test_decompress_z80v2(self):
        exp_ram = [33] * 263 + [7] * 11 + [144] * 3 + [13, 14] + [237] * 3
        exp_ram += [107] * (49152 - len(exp_ram))
        self._test_z80(exp_ram, 2, True)

    def test_z80v3(self):
        exp_ram = [(n + 128) & 255 for n in range(49152)]
        self._test_z80(exp_ram, 3, False)

    def test_decompress_z80v3(self):
        exp_ram = [201] * 4336 + [19] * 255 + [142] * 2 + [1, 2] + [37] * 4
        exp_ram += [193] * (49152 - len(exp_ram))
        self._test_z80(exp_ram, 3, True)

    def test_sna(self):
        header = [0] * 27
        exp_ram = [(n + 23) & 255 for n in range(49152)]
        sna = header + exp_ram
        tmp_sna = self.write_bin_file(sna, suffix='.sna')
        snapshot = get_snapshot(tmp_sna)
        ram = snapshot[16384:]
        self.assertEqual(len(ram), 49152)
        self.assertEqual(ram, exp_ram)

    def test_unknown_file_type(self):
        file_type = 'szx'
        snapshot_file = self.write_bin_file(suffix='.{0}'.format(file_type))
        try:
            get_snapshot(snapshot_file)
            self.fail()
        except SnapshotError as e:
            self.assertEqual(e.args[0], "{0}: Unknown file type '{1}'".format(snapshot_file, file_type))

    def test_bad_ram_size(self):
        for ram_size in (3, 49200):
            try:
                get_snapshot(self.write_bin_file([0] * (27 + ram_size), suffix='.sna'))
                self.fail()
            except SnapshotError as e:
                self.assertEqual(e.args[0], 'RAM size is {0}'.format(ram_size))

    def test_bad_z80(self):
        header = [0] * 30
        header[6] = 255 # Set PC > 0 to indicate a v1 Z80 snapshot
        header[12] |= 32 # Signal that the RAM data block is compressed
        z80 = header + [255] # Good byte to start with
        z80 += [237, 237, 0, 11] # Bad block of length 0
        z80 += [0, 237, 237, 0] # Terminator
        z80_file = self.write_bin_file(z80, suffix='.z80')
        try:
            get_snapshot(z80_file)
            self.fail()
        except SnapshotError as e:
            self.assertEqual(e.args[0], 'Found ED ED 00 0B')

if __name__ == '__main__':
    unittest.main()
